from httpx import URL, Cookies, Headers, QueryParams, Request, Response
from httpx._auth import AuthTypes
from httpx._content_streams import RequestData, RequestFiles
from httpx._models import CookieTypes, HeaderTypes, QueryParamTypes

__all__ = [
    "URL",
    "Cookies",
    "Headers",
    "QueryParams",
    "Request",
    "Response",
    "AuthTypes",
    "RequestData",
    "RequestFiles",
    "CookieTypes",
    "HeaderTypes",
    "QueryParamTypes",
]



from collections import abc
from enum import Enum
from inspect import isclass
from typing import Any, Optional, Type, Union

from .compat import get_args, get_origin
from .nodes import *
from .reporting import ErrorReporter, LogErrorReporter, PrettyJson5Formatter


class Fitter:
    """
    Core class responsible for the fitting of objects.

    - Create an instance with the configuration you want
    - Use the :py:meth:`~.fit` method to do your fittings

    Notes
    -----
    Overall orchestrator of the fitting. A lot of the logic happens in the
    nodes, but this class is responsible for executing the logic in the right
    order and also holds configuration.
    """

    def __init__(
        self,
        no_unwanted_keys: bool = False,
        error_reporter: Optional[ErrorReporter] = None,
    ):
        """
        Constructs the instance.

        Parameters
        ----------
        no_unwanted_keys
            If set to ``True``, it will not be allowed to have unwanted keys
            when fitting mappings into dataclasses/named tuples.
        error_reporter
            Error reporting for when a validation fails. By default no report
            is made but you might want to arrange reporting for your needs,
            otherwise you're going to be debugging in the blind.
        """

        self.no_unwanted_keys = no_unwanted_keys
        self.error_reporter = error_reporter

    def _as_node(self, value: Any):
        """
        Recursively transforms a value into a node.

        Parameters
        ----------
        value
            Any kind of JSON-decoded value (string, list, object, etc).
        """

        if isinstance(value, (int, float, str, bool)) or value is None:
            return FlatNode(self, value)
        elif isinstance(value, abc.Sequence):
            return ListNode(self, value, [self._as_node(x) for x in value])
        elif isinstance(value, abc.Mapping):
            return MappingNode(
                self, value, {k: self._as_node(v) for k, v in value.items()}
            )
        else:
            raise ValueError

    def _fit_union(self, t: Type[T], value: Node) -> T:
        """
        In case of a union, walk through all possible types and try them on
        until one fits (fails otherwise).
        """

        for sub_t in get_args(t):
            try:
                return self.fit_node(sub_t, value)
            except ValueError:
                continue

        value.fail("No matching type in Union")

    def _fit_class(self, t: Type[T], value: Node) -> T:
        """
        Wrapper around the ``FlatNode``'s fit method.
        """

        if isinstance(value, FlatNode):
            return value.fit(t)

        value.fail(f"Node is not {t}")

    def _fit_any(self, _: Type[T], value: Node) -> T:
        """
        That's here for consistency but let's be honest, this function is not
        very useful.
        """

        return value.value

    def _fit_none(self, _: Type[T], value: Node) -> T:
        """
        Does basic checks before returning None or raising an error
        """

        if value.value is None:
            value.fit_success = True
            return None

        value.fail(f"Value is not None")

    def _fit_enum(self, t: Type[T], value: Node) -> T:
        """
        Tries to find back the right enum value
        """

        try:
            out = t(value.value)
        except ValueError:
            value.fail(f"No match in enum {t!r}")
        else:
            value.fit_success = True
            return out

    def fit_node(self, t: Type[T], value: Node) -> T:
        """
        Tries to find the right fit according to the type you're trying to
        match and the node type.

        Notes
        -----
        The order of tests done in this code is very important. By example,
        dataclasses would pass the ``isclass(t)`` test so ``MappingNode`` have
        to be handled before that test.

        Parameters
        ----------
        t
            Type you want to fit your node into
        value
            A node you want to fit into a type

        Raises
        ------
        ValueError
        """

        if get_origin(t) is Union:
            return self._fit_union(t, value)
        elif t is Any:
            return self._fit_any(t, value)
        elif isinstance(value, (MappingNode, ListNode)):
            return value.fit(t)
        elif t is None or t is None.__class__:
            return self._fit_none(t, value)
        elif isclass(t):
            if issubclass(t, Enum):
                return self._fit_enum(t, value)
            else:
                return self._fit_class(t, value)
        else:
            value.fail("Could not fit. This error can never be reached in theory.")

    def fit(self, t: Type[T], value: Any) -> T:
        """
        Fits data into a type. The data is expected to be JSON-decoded values
        (strings, ints, bools, etc).

        On failure a ValueError will arise and if an error reporter is set it
        will be sent the node to generate the error report.

        Parameters
        ----------
        t
            Type you want to fit the value into
        value
            Value you want to fit into a type

        Raises
        -------
        ValueError
        """

        node = self._as_node(value)

        try:
            return self.fit_node(t, node)
        except ValueError:
            if self.error_reporter:
                self.error_reporter.report(node)
            raise


def typefit(t: Type[T], value: Any) -> T:
    """
    Fits a JSON-decoded value into native Python type-annotated objects.

    This uses the default sane settings but it might not be up to your taste.
    By example, errors will be reported in the logging module using ANSI escape
    codes for syntactical coloration, however depending on the situation you
    might not want that.

    If you want more flexibility and configuration, you can use the
    :py:class:`~.Fitter` directly.

    Parameters
    ----------
    t
        Type to fit the value into. Currently supported types are:

          - Simple builtins like :class:`int`, :class:`float`,
            :class:`typing.Text`, :class:`typing.bool`
          - Enumerations which are subclass of :class:`enum.Enum`.
          - Custom types. The constructor needs to accept exactly one parameter
            and that parameter should have a typing annotation.
          - :class:`typing.Union` to define several possible types
          - :class:`typing.List` to declare a list and the type of list values
    value
        Value to be fit into the type

    Returns
    -------
    T
        If the value fits, a value of the right type is returned.

    Raises
    ------
    ValueError
        When the fitting cannot be done, a :class:`ValueError` is raised.

    See Also
    --------
    Fitter.fit
    """

    return Fitter(
        error_reporter=LogErrorReporter(
            formatter=PrettyJson5Formatter(colors="terminal16m")
        )
    ).fit(t, value)



"""
Compatibility module that backports the get_origin and get_args functions from
Python 3. This code is from the official standard library.

Copyright © 2001-2019 Python Software Foundation; All Rights Reserved
"""

import collections

# noinspection PyUnresolvedReferences,PyProtectedMember
from typing import Any, Generic, Tuple, _GenericAlias

try:
    from typing import get_origin
except ImportError:

    def get_origin(tp) -> Any:
        """Get the unsubscripted version of a type.

        This supports generic types, Callable, Tuple, Union, Literal, Final and ClassVar.
        Return None for unsupported types. Examples::

            get_origin(Literal[42]) is Literal
            get_origin(int) is None
            get_origin(ClassVar[int]) is ClassVar
            get_origin(Generic) is Generic
            get_origin(Generic[T]) is Generic
            get_origin(Union[T, int]) is Union
            get_origin(List[Tuple[T, T]][int]) == list
        """
        if isinstance(tp, _GenericAlias):
            return tp.__origin__
        if tp is Generic:
            return Generic
        return None


try:
    from typing import get_args  # lgtm[py/unused-import]
except ImportError:

    def get_args(tp) -> Tuple[Any, ...]:
        """Get type arguments with all substitutions performed.

        For unions, basic simplifications used by Union constructor are performed.
        Examples::
            get_args(Dict[str, int]) == (str, int)
            get_args(int) == ()
            get_args(Union[int, Union[T, int], str][int]) == (int, str)
            get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
            get_args(Callable[[], T][int]) == ([], int)
        """
        if isinstance(tp, _GenericAlias):
            res = tp.__args__
            if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        return ()



import re
from collections import abc
from inspect import Parameter, isclass, signature
from string import Formatter
from typing import Any, Callable, Dict, Generic, Iterator, Text, TypeVar
from urllib.parse import quote_plus

CLASS_RE = re.compile(r"<class '([^']+)'>")

T = TypeVar("T")


def loose_call(func: Callable, kwargs: Dict[Text, Any]):
    """
    Calls a function using only kwargs and drops extra parameters that are not
    required if there is no kwargs argument to collect extra arguments.
    """

    sig = signature(func)

    has_kwargs = False

    for param in sig.parameters.values():
        if param.kind == Parameter.VAR_KEYWORD:
            has_kwargs = True

    if not has_kwargs:
        expect = set(sig.parameters.keys())
        present = set(kwargs.keys())

        return func(**{k: kwargs[k] for k in (expect & present)})
    else:
        return func(**kwargs)


def callable_value(value, kwargs):
    """
    Used for optional callable options in the API generator.
    """

    if callable(value):
        return loose_call(value, kwargs)
    return value


class UrlFormatter(Formatter):
    """
    Just like a regular formatter except that all formatted variables are
    escaped to be URL-safe. It allows to create easily paths like
    @get("items/{item_id}.json") without worrying about escaping the ID.
    """

    def format_field(self, value, format_spec):
        out = super().format_field(value, format_spec)
        out = quote_plus(f"{out}")
        return out


def is_named_tuple(value: Any) -> bool:
    if isclass(value):
        test = issubclass
    else:
        test = isinstance

    return test(value, tuple) and hasattr(value, "_fields")


class OrderedSet(Generic[T], abc.MutableSet):
    """
    Behaves exactly like a set() except that the objects will be kept in order
    of insertion.

    Notes
    -----
    Internally it relies on dictionaries keeping order. This means that this
    will only work for Python 3.7+ (and CPython 3.6+).
    """

    def __init__(self, initial_data: Iterator[T] = tuple()):
        self._set = {k: True for k in initial_data}

    def add(self, x: T) -> None:
        self._set[x] = True

    def discard(self, x: T) -> None:
        try:
            del self._set[x]
        except KeyError:
            pass

    def __contains__(self, x: object) -> bool:
        return x in self._set

    def __len__(self) -> int:
        return len(self._set)

    def __iter__(self) -> Iterator[T]:
        yield from self._set


def format_type_name(t: Any) -> Text:
    """
    Heuristics to make a type name look nice. It might change in the future.

    Parameters
    ----------
    t
        Type whose name you want to format
    """

    out = f"{t}"

    m = CLASS_RE.match(out)

    if m:
        out = f"'{m.group(1)}'"

    if out == "'NoneType'":
        out = "'None'"

    return out



from collections import ChainMap, abc
from dataclasses import fields, is_dataclass
from datetime import date, datetime
from enum import Enum
from json import dumps
from typing import Any, get_type_hints
from uuid import UUID

from .meta import Source


class Serializer:
    """
    Base serializer, that has no opinion and will serialize anything that is
    100% to serialize without making any assumption.

    Supported types are:

    - Numbers (int, float)
    - Booleans
    - Strings
    - Sequences
    - Mappings
    - Named (and typed) tuples
    - Dataclasses (including with Typefit metadata in the field)
    - Any object with a `__typefit_serialize__()` method
    - Enums

    You'll notice that the behavior of this class is a best effort to make
    something sane and simple. This means there is no warranty that this works:

    >>> base: SomeData = # Some data
    >>> serialized = serialize(base)
    >>> assert typefit(SomeData, serialized) == base

    If you want more types to be recognized by this serializer, you can inherit
    from it and extend the :py:meth:`~.Serializer.find_serializer()` method.

    If you don't know where to look, check out the following methods:

    - :py:meth:`~.typefit.serialize.Serializer.serialize`
    - :py:meth:`~.typefit.serialize.Serializer.json`

    See Also
    --------
    SaneSerializer
    """

    def find_serializer(self, obj: Any):
        """
        Trying to be as generic as possible. There is a few tricks there, like
        strings which are also sequences so the order of tests matters quite a
        lot.

        Please override this if you want to change the behavior. See how it's
        done in :py:class:`~.typefit.serialize.SaneSerializer` for an idea
        on how to do it.
        """

        if hasattr(obj, "__typefit_serialize__"):
            return self.serialize_typefit
        elif isinstance(obj, (int, float, bool, str)) or obj is None:
            return self.serialize_generic
        elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
            return self.serialize_tuple
        elif is_dataclass(obj):
            return self.serialize_dataclass
        elif isinstance(obj, abc.Sequence):
            return self.serialize_sequence
        elif isinstance(obj, abc.Mapping):
            return self.serialize_mapping
        elif isinstance(obj, Enum):
            return self.serialize_enum

    def serialize_generic(self, obj: Any) -> Any:
        """
        By default, leave the object untouched
        """

        return obj

    def serialize_tuple(self, obj: tuple):
        """
        Named tuples are expected to have typing annotations, we'll use that
        as a reference to get the fields list, however types are not enforced.
        """

        return {
            k: self.serialize(getattr(obj, k)) for k in get_type_hints(obj.__class__)
        }

    def serialize_sequence(self, obj: abc.Sequence):
        """
        Sequences are converted to regular lists, and each item of the list
        is recursively serialized.
        """

        return [self.serialize(x) for x in obj]

    def serialize_typefit(self, obj: Any):
        """
        Serializes an object by calling its `__typefit_serialize__()` method.
        """

        return obj.__typefit_serialize__()

    def serialize_dataclass(self, obj: Any):
        """
        Dataclasses are mappings but they merit a special attention due to the
        fact that their fields are not necessarily the fields that will be
        used in the output, thanks to the `meta(source=X)` feature.

        Notes
        -----
        See :py:class:`~.typefit.meta.Source`, but basically the conversion to
        JSON structure generates a series of dictionaries that are then
        superposed into a single dictionary and returned.

        All values of this dictionary are of course recursively serialized.
        """

        def _get_values():
            for field in fields(obj.__class__):
                if field.metadata and "typefit_source" in field.metadata:
                    source: Source = field.metadata["typefit_source"]
                    yield {
                        k: self.serialize(v)
                        for k, v in source.value_to_json(field.name, obj).items()
                    }
                else:
                    yield {field.name: self.serialize(getattr(obj, field.name))}

        return dict(ChainMap(*_get_values()))

    def serialize_mapping(self, obj: abc.Mapping):
        """
        Mappings are just copied into another mapping. While copying, all the
        values are recursively serialized.
        """

        return {k: self.serialize(v) for k, v in obj.items()}

    def serialize_enum(self, obj: Enum):
        """
        Enums are serialized into their value.
        """

        return self.serialize(obj.value)

    def serialize(self, obj: Any):
        """
        Transforms a given object into a structure of basic types which can
        easily be serialized to JSON. It is not a strict inverse of
        :py:func:`~.typefit.typefit` but it should be good enough for most
        uses.

        Please note that this at least assumes that objects are consistent with
        their type declarations, no additional security is put in place.

        This method relies on the :py:meth:`~.Serializer.find_serializer()`
        method, which means that if you implement a subclass in order to
        change the mapping of serialization functions you should override
        :py:meth:`~.Serializer.find_serializer()`.

        Parameters
        ----------
        obj
            Object to be serialized
        """

        serializer = self.find_serializer(obj)
        return serializer(obj)

    def json(self, obj: Any) -> str:
        """
        Shortcut to transform an object into a JSON string going through
        :py:meth:`~.serialize`.
        """

        return dumps(self.serialize(obj))


class SaneSerializer(Serializer):
    """
    Opinionated version of what sane default for non-JSON-standard types
    should be. Comes as an extension of
    :py:class:`~.Serializer`.

    - Dates are serialized to the ISO format
    - UUIDs are serialized into their default str() representation
    """

    def find_serializer(self, obj: Any):
        """
        Tries to find special cases and if none of them are matched then
        resort to the parent method.
        """

        if isinstance(obj, datetime):
            return self.serialize_std_datetime
        elif isinstance(obj, date):
            return self.serialize_std_date
        elif isinstance(obj, UUID):
            return self.serialize_uuid
        else:
            return super().find_serializer(obj)

    def serialize_uuid(self, obj: UUID):
        """
        UUIDs are simply converted to strings
        """

        return f"{obj}"

    def serialize_std_datetime(self, obj: datetime):
        """
        Datetime are converted into ISO format
        """

        return obj.isoformat()

    def serialize_std_date(self, obj: date):
        """
        Dates are converted to ISO format
        """

        return obj.isoformat()


def serialize(obj: Any) -> Any:
    """
    Shortcut to use the :py:class:`~.typefit.serialize.SaneSerializer`'s
    :py:meth:`~.typefit.serialize.Serializer.serialize` method.

    Parameters
    ----------
    obj
        Object to be serializer
    """

    return SaneSerializer().serialize(obj)



from functools import wraps
from inspect import signature
from typing import Any, Callable, Dict, List, Optional, Text, Type, Union
from urllib.parse import urljoin

import httpx

from . import httpx_models as hm
from .fitting import T, typefit
from .serialize import SaneSerializer
from .utils import UrlFormatter, callable_value

HeadersFactory = Callable[..., hm.HeaderTypes]
Headers = Union[None, hm.HeaderTypes, HeadersFactory]

PathFactory = Callable[..., Text]
Path = Union[PathFactory, Text]

ParamsFactory = Callable[..., hm.QueryParamTypes]
Params = Union[None, hm.QueryParamTypes, ParamsFactory]

CookiesFactory = Callable[..., hm.CookieTypes]
Cookies = Union[None, hm.CookieTypes, CookiesFactory]

AuthFactory = Callable[..., hm.AuthTypes]
Auth = Union[None, hm.AuthTypes, AuthFactory]

AllowRedirectsFactory = Callable[..., bool]
AllowRedirects = Union[None, bool, AllowRedirectsFactory]

DataFactory = Callable[..., hm.RequestData]
Data = Union[None, hm.RequestData, DataFactory]

FilesFactory = Callable[..., hm.RequestFiles]
Files = Union[None, hm.RequestFiles, FilesFactory]

JsonType = Union[Dict[Text, "JsonType"], List["JsonType"], int, float, bool, Text, None]
JsonFactory = Callable[..., hm.RequestFiles]
Json = Union[None, JsonType, JsonFactory]


def _make_decorator(
    method: Text,
    path: Path,
    data: Data = None,
    files: Files = None,
    json: Json = None,
    params: Params = None,
    headers: Headers = None,
    cookies: Cookies = None,
    auth: Auth = None,
    allow_redirects: AllowRedirects = None,
    hint: Any = None,
) -> Callable[[Callable], Callable]:
    """
    Generates a decorator that can be used for any kind of HTTP request.
    That's a bit messy but without it the GET/POST/etc decorators would share
    very duplicated code.
    """

    def decorator(func: Callable):
        sig = signature(func)

        if len(sig.parameters) == 0:
            raise TypeError("Decorated function doesn't have any arguments")

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            Gets the signature from the decorated function and applies the
            arguments received from the call. If the arguments are no good then
            it will fail.

            Then uses the helper's get method in order to generate the actual
            request and arguments.

            The real method is never called.
            """

            bound = sig.bind(*args, **kwargs)
            bound.apply_defaults()

            self = next(iter(bound.arguments.values()))

            if not isinstance(self, SyncClient):
                raise TypeError(f"{self!r} is not a SyncClient")

            return self.helper.request(
                method=method,
                path=path,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                cookies=cookies,
                auth=auth,
                allow_redirects=allow_redirects,
                hint=hint,
                kwargs=bound.arguments,
                data_type=sig.return_annotation,
            )

        return wrapper

    return decorator


def get(
    path: Path,
    params: Params = None,
    headers: Headers = None,
    cookies: Cookies = None,
    auth: Auth = None,
    allow_redirects: AllowRedirects = None,
    hint: Any = None,
):
    """
    Generates an API method that GET the URL, based on provided parameters and
    method signature. The decorated method's code will never be called, only
    the generated method will be used.
    """

    return _make_decorator(
        "get",
        path=path,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        allow_redirects=allow_redirects,
        hint=hint,
    )


def post(
    path: Path,
    data: Data = None,
    files: Files = None,
    json: Json = None,
    params: Params = None,
    headers: Headers = None,
    cookies: Cookies = None,
    auth: Auth = None,
    allow_redirects: AllowRedirects = None,
    hint: Any = None,
):
    """
    Generates an API method that POST the URL, based on provided parameters and
    method signature. The decorated method's code will never be called, only
    the generated method will be used.
    """

    return _make_decorator(
        "post",
        path=path,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        allow_redirects=allow_redirects,
        hint=hint,
    )


def put(
    path: Path,
    data: Data = None,
    files: Files = None,
    json: Json = None,
    params: Params = None,
    headers: Headers = None,
    cookies: Cookies = None,
    auth: Auth = None,
    allow_redirects: AllowRedirects = None,
    hint: Any = None,
):
    """
    Generates an API method that PUT the URL, based on provided parameters and
    method signature. The decorated method's code will never be called, only
    the generated method will be used.
    """

    return _make_decorator(
        "put",
        path=path,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        allow_redirects=allow_redirects,
        hint=hint,
    )


def patch(
    path: Path,
    data: Data = None,
    files: Files = None,
    json: Json = None,
    params: Params = None,
    headers: Headers = None,
    cookies: Cookies = None,
    auth: Auth = None,
    allow_redirects: AllowRedirects = None,
    hint: Any = None,
):
    """
    Generates an API method that PATCH the URL, based on provided parameters
    and method signature. The decorated method's code will never be called,
    only the generated method will be used.
    """

    return _make_decorator(
        "patch",
        path=path,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        auth=auth,
        allow_redirects=allow_redirects,
        hint=hint,
    )


class _SyncClientHelper:
    """
    Effector for all requests and parameters generation. It's separated from
    the client itself for readability regarding what should or should not be
    overridden.
    """

    def __init__(self, client: "SyncClient"):
        self.client = client
        self.http = httpx.Client()

    def close(self):
        """
        Closes the underlying HTTP connection pool
        """

        self.http.close()

    def url(self, path: Path, kwargs: Dict[Text, Any]):
        """
        Generates the URL using urljoin in the client's BASE_URL and the
        provided path. The path could be a callable, if so it will be called
        using loose_call and the provided kwargs.
        """

        f = UrlFormatter()
        url = urljoin(self.client.BASE_URL, callable_value(path, kwargs))
        return f.format(url, **kwargs)

    def headers(self, extra: Headers, kwargs: Dict[Text, Any]) -> hm.Headers:
        """
        Generates the headers for this request. It will:

        1. Get the default headers as generated by the client's headers()
           method
        2. Use the specific extra headers specified via the decorator. Callable
           values will be called before being returned.

        All of that is merged and returned.
        """

        out = hm.Headers(self.client.headers())
        out.update(callable_value(extra, kwargs))

        return out

    def cookies(self, extra: Cookies, kwargs: Dict[Text, Any]) -> hm.Cookies:
        """
        Generates the cookies for this request. It will:

        1. Get the default cookies as generated by the client's cookies()
           method
        2. Use the specific extra cookies specified via the decorator. Callable
           values will be called before being returned.

        All of that is merged and returned.
        """

        out = hm.Cookies(self.client.cookies())
        out.update(callable_value(extra, kwargs))

        return out

    def auth(self, override: Auth, kwargs: Dict[Text, Any]) -> Auth:
        """
        If there is an override from the decorator then this prevails over the
        static auth provided by the client class but otherwise it will just use
        the output of the auth() method in the client.
        """

        ov = callable_value(override, kwargs)

        if ov:
            return ov

        return self.client.auth()

    def allow_redirects(
        self, override: AllowRedirects, kwargs: Dict[Text, Any]
    ) -> bool:
        """
        Checks if the decorator attempts an override (by returning a non-None
        value), otherwise stick to the client's value.
        """

        ov = callable_value(override, kwargs)

        if ov is not None:
            return ov

        return self.client.allow_redirects()

    def request(
        self,
        method: Text,
        kwargs: Dict[Text, Any],
        data_type: Type[T],
        path: Text,
        data: Data = None,
        files: Files = None,
        json: Json = None,
        headers: Headers = None,
        cookies: Cookies = None,
        auth: Auth = None,
        allow_redirects: AllowRedirects = None,
        params: Params = None,
        hint: Any = None,
    ) -> T:
        """
        This will generate a call to HTTPX depending on the provided overrides
        (in the arguments) and available default values as declared by the
        client. Arguments will be selected automatically depending on the
        method.
        """

        request_args = dict(
            url=self.url(path, kwargs),
            headers=self.headers(headers, kwargs),
            params=callable_value(params, kwargs),
            cookies=self.cookies(cookies, kwargs),
            auth=self.auth(auth, kwargs),
            allow_redirects=self.allow_redirects(allow_redirects, kwargs),
        )

        if method in {"post", "put", "patch"}:
            request_args.update(
                data=callable_value(data, kwargs),
                files=callable_value(files, kwargs),
                json=self.client.serialize(callable_value(json, kwargs)),
            )

        r = getattr(self.http, method)(**request_args)
        self.client.raise_errors(r, hint)
        data = self.client.decode(r, hint)
        data = self.client.extract(data, hint)

        return self.client.typefit(data_type, data)


class SyncClient:
    """
    SyncClient base class. To create your own API client, inherit from this
    and generate your methods using the HTTP decorators found above.
    """

    BASE_URL = ""

    def __init__(self):
        self.helper = _SyncClientHelper(self)
        self.serialize = self.init_serialize()
        self.typefit = self.init_typefit()

    def init_typefit(self) -> Callable[[Type[T], Any], T]:
        """
        Uses :py:func:`~.typefit.typefit` by default, however you might want to
        configure things differently, like the logging of errors. If you want
        to do so, just override this method.
        """

        return typefit

    def init_serialize(self) -> Callable[[Any], Any]:
        """
        Everything going through the `json` parameter will be serialized using
        the serializer function returned by this method. By default this
        uses the :py:class:`~.typefit.serialize.SaneSerializer` class which
        is helpful in the general case. If you want something more specific,
        feel free to override this method.

        If you don't want your data to be tampered with at all, you can do
        something like

        >>> def init_serialize(self):
        >>>     return lambda x: x
        """

        return SaneSerializer().serialize

    def close(self):
        """
        Closes the underlying HTTP connection
        """

        self.helper.close()

    def headers(self) -> Optional[hm.HeaderTypes]:
        """
        Inherit this to generate headers that will be sent at each request.
        """

    def cookies(self) -> Optional[hm.CookieTypes]:
        """
        Inherit this to generate cookies to be sent at each request
        """

    def auth(self) -> Optional[hm.AuthTypes]:
        """
        Inherit this to generate auth to be sent at each request
        """

    def allow_redirects(self) -> bool:
        """
        Return False to disable redirects. Also, if a value is specified in the
        decorator then this value will be overridden.
        """

        return True

    def raise_errors(self, resp: httpx.Response, hint: Any) -> None:
        """
        By default, raise errors if HTTP statuses are error status but you
        could do any kind of inspection you want here.

        The hint is there in case you need different mechanisms for different
        paths, the hint is provided through the decorator.
        """

        resp.raise_for_status()

    def decode(self, resp: httpx.Response, hint: Any) -> Any:
        """
        Transforms the HTTP response into viable data. By default it decodes
        JSON but who knows what you might want to support.

        The hint is there in case you need different mechanisms for different
        paths, the hint is provided through the decorator.
        """

        return resp.json()

    def extract(self, data: Any, hint: Any) -> Any:
        """
        Use this method to extract the data before fitting it into a type. By
        example, sometimes APIs will return something like {result: xxx}
        instead of returning your object directly.

        The hint is there in case you need different mechanisms for different
        paths, the hint is provided through the decorator.
        """

        return data



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from logging import ERROR, getLogger
from typing import Iterator, List, Text

from pygments import highlight
from pygments.formatters import get_formatter_by_name
from pygments.lexers.javascript import JavascriptLexer

from .nodes import *

logger = getLogger("typefit")


class ErrorFormatter(ABC):
    """
    This interface is in charge of converting the meta information from
    :py:class:`~.typefit.nodes.Node` into some readable text report. That's
    useful for some reporters, like the :py:class:`~.LogErrorReporter`.
    """

    @abstractmethod
    def format(self, node: "Node") -> Text:
        """
        Implement this in order to transform the node into some readable text.

        Parameters
        ----------
        node
            Node that failed to be parsed, with errors
        """

        raise NotImplementedError


class ErrorReporter(ABC):
    """
    Implement this interface to be able to report errors. Once configured into
    the :py:class:`~.typefit.Fitter`, every time a fitting fails this will be
    triggered with the node that failed to be converted so that the error can
    be reported to the developer for debugging purposes.

    You can imagine reporters that send the error to the logging module
    (see :py:class:`~.LogErrorReporter`) but it could as well be displaying
    the data in an interactive web page or an IDE plugin to help seeing clearly
    what is happening.
    """

    @abstractmethod
    def report(self, node: "Node") -> None:
        """
        Will be called by :py:class:`~.typefit.Fitter` for every error.
        Override it in order to report the error wherever you want to report it
        to.

        Parameters
        ----------
        node
            Node that failed to fit
        """

        raise NotImplementedError


class LogErrorReporter(ErrorReporter):
    """
    Reports the errors into the logging module

    Parameters
    ----------
    formatter
        Formatter to use
    level
        Log level to report the errors with
    """

    def __init__(self, formatter: ErrorFormatter, level: int = ERROR):
        self.formatter = formatter
        self.level = level

    def report(self, node: "Node") -> None:

        logger.log(self.level, "Fitting error:\n\n%s", self.formatter.format(node))


@dataclass
class _Line:
    """
    Internal utility class to represent a line that might have errors above
    its head.

    - ``level`` -- Indent level
    - ``content`` -- Content of the line
    - ``errors`` -- List of errors attributed to that line
    """

    level: int
    content: Text
    errors: List[Text] = field(default_factory=list)


class PrettyJson5Formatter(ErrorFormatter):
    """
    Formats the data into annotated and redacted JSON5. The goal here is to
    display the errors on top of the actual data, while removing all the data
    that is not useful to understand the error itself.

    The output can be colored by Pygments, just fill in the ``colors`` argument
    with a valid Pygments formatter name (could be ANSI codes, could be HTML,
    could be whatever you desire).

    Parameters
    ----------
    indent
        Content of one ident level (default = 4 spaces)
    colors
        Pygments color formatter. No coloring happens if this string is
        empty.
    truncate_strings_at
        If a string is longer than this indicated length then it will be
        truncated for display. Default is 40, put 0 to disable the feature.
    """

    def __init__(
        self, indent: Text = "    ", colors: Text = "", truncate_strings_at: int = 40,
    ):
        self.indent = indent
        self.colors = colors
        self.truncate_strings_at = truncate_strings_at
        self._previous_level = -1

    def _indent(self, level: int) -> Text:
        """
        Generates the indent for the given indent level
        """

        return self.indent * level

    def _line(self, line: _Line) -> Text:
        """
        Generates the output for a given line.

        Notes
        -----
        There is some fancy logic here to make sure that if there is errors
        on the line and if it's not the first line at this indent level then
        we should leave a blank line above. This is to make it clear which line
        the error comments are attached to.

        Parameters
        ----------
        line
            Line to print
        """

        out = "\n".join(
            [
                *([""] if line.errors and line.level == self._previous_level else []),
                *(f"{self._indent(line.level)}// {e}" for e in line.errors),
                f"{self._indent(line.level)}{line.content}",
            ]
        )

        self._previous_level = line.level

        return out

    def _format_flat(
        self, node: FlatNode, level: int, prefix: Text, suffix: Text
    ) -> Iterator[_Line]:
        """
        Formats a flat literal using the standard JSON function.

        Notes
        -----
        String-truncating logic happens here, that's why this function is a
        little bit more complicated.
        """

        if (
            isinstance(node.value, str)
            and self.truncate_strings_at
            and len(node.value) > self.truncate_strings_at
        ):
            value = f"{node.value[:self.truncate_strings_at]}[...]"
            extra = " /* truncated */"
        else:
            value = node.value
            extra = ""

        out = _Line(
            level, f"{prefix}{json.dumps(value, ensure_ascii=False)}{extra}{suffix}"
        )

        if not node.fit_success:
            out.errors.extend(node.errors)

        yield out

    def _format_list(
        self, node: ListNode, level: int, prefix: Text, suffix: Text
    ) -> Iterator[_Line]:
        """
        Formats a list. Same logic as :py:meth:`~._format_mapping` except for
        lists.
        """

        if not node.problem_is_kids:
            if node.value:
                content = " /* ... */ "
            else:
                content = ""

            yield _Line(level, f"{prefix}[{content}]{suffix}", [*node.errors])
        else:
            yield _Line(level, f"{prefix}[", [*node.errors])

            errors = set()
            to_display = []

            for child in node.children:
                child_errors = tuple(child.errors)

                if not child.fit_success and child_errors not in errors:
                    errors.add(child_errors)
                    to_display.append(child)

            for child in to_display:
                yield from self._format(child, level + 1, "", ",")

            yield _Line(level, f"]{suffix}")

    def _format_mapping(
        self, node: MappingNode, level: int, prefix: Text, suffix: Text
    ) -> Iterator[_Line]:
        """
        Formats a mapping.

        - If the problem is not in the kids (maybe an int was expected and an
          object was received, by example) then just abreviate the object and
          put the error on top
        - If the problem is with the kids, recursively renders the kids to
          help the developer figuring out which kid is misbehaving
        """

        if not node.problem_is_kids:
            if node.value:
                content = " /* ... */ "
            else:
                content = ""

            yield _Line(level, f"{prefix}{{{content}}}{suffix}", [*node.errors])
        else:
            yield _Line(level, f"{prefix}{{", [*node.errors])

            for key, child in node.children.items():
                yield from self._format(child, level + 1, f"{json.dumps(key)}: ", ",")

            if node.missing_keys:
                yield _Line(
                    level + 1,
                    f'// Missing keys: {", ".join(repr(k) for k in node.missing_keys)}',
                )

            yield _Line(level, f"}}{suffix}")

    def _format(
        self, node: Node, level: int = 0, prefix: Text = "", suffix: Text = ""
    ) -> Iterator[_Line]:
        """
        Will recursively generate the format of a line. Depending on the node
        type, the right formatter will be chosen and all the generated lines
        will be iterated.

        Notes
        -----
        Here there is a prefix/suffix logic here which allows to easily put
        a coma after the line or a key name in front of the line without
        knowing beforehand where the line is going to be displayed or if it's
        the first/last line.

        Parameters
        ----------
        node
            Node to format
        level
            Indent level
        prefix
            Prefix to put before the line (on the first line)
        suffix
            Suffix to put after the line (on the last line)

        See Also
        --------
        _format_flat
        _format_list
        _format_mapping
        """

        # noinspection PyTypeChecker
        yield from {
            FlatNode: self._format_flat,
            ListNode: self._format_list,
            MappingNode: self._format_mapping,
        }[node.__class__](node, level, prefix, suffix)

    def format(self, node: "Node") -> Text:
        """
        Formats the node into JSON5. If colors were specified in the
        constructor, then it's also where the coloration is added before being
        returned.
        """

        out = "\n".join(map(self._line, self._format(node)))
        formatter = None

        if self.colors:
            formatter = get_formatter_by_name(self.colors)

        if formatter:
            out = highlight(out, JavascriptLexer(), formatter)

        return out



from dataclasses import dataclass, field, fields, is_dataclass
from inspect import Parameter, isclass, signature
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Text,
    Type,
    TypeVar,
    get_type_hints,
)

from .compat import get_args, get_origin
from .meta import Source
from .utils import OrderedSet, format_type_name, is_named_tuple

if TYPE_CHECKING:
    from .fitting import Fitter


__all__ = ["T", "Node", "MappingNode", "ListNode", "FlatNode"]


T = TypeVar("T")


@dataclass
class Node:
    """
    Typefit is made to convert JSON structure into Python dataclasses and
    objects. In order to make analyzing a fitting possible after it failed,
    data structures will be converted into a tree of Nodes. This is the base
    class, but there is 3 kinds of nodes:

    - :py:class:`~.FlatNode` -- "Flat" structures, aka numbers, strings, bools
      or nulls
    - :py:class:`~.MappingNode` -- For mappings (dictionaries, dataclasses and
      named tuples)
    - :py:class:`~.ListNode` -- For lists

    Each node will remember the errors that the decoder went through and if
    the decoding was successful at all. In other terms, by navigating through
    the nodes you can generate readable error reporting.

    Other Parameters
    ----------------
    fitter
        Fitter instance to refer to for configuration and fitting other nodes
    value
        Value of this node
    errors
        All unique errors encountered on this node. There might still be errors
        if the fitting was successful, as those errors record all attempts.
    fit_success
        Indicates if the fitting was successful at least once. This does not
        mean that the overall fitting went well and especially with unions this
        could be tricky.

    See Also
    --------
    FlatNode
    MappingNode
    ListNode
    """

    fitter: "Fitter" = field(repr=False)
    value: Any
    errors: Set[Text] = field(init=False, default_factory=OrderedSet, repr=False)
    fit_success: bool = field(init=False, default=False)

    def fit(self, t: Type[T]) -> T:
        """
        That is left to the subclasses to implement
        """

        raise NotImplementedError

    def fail(self, reason: Text) -> None:
        """
        Utility to trigger a failure

        - Register the error locally.
        - Raise a value error. Nodes that have children nodes need to catch the
          value errors from their children in order to not stop their own
          processing.

        Parameters
        ----------
        reason
            Reason to be given and registered as error
        """

        self.errors.add(reason)
        raise ValueError(reason)


@dataclass
class MappingNode(Node):
    """
    Those nodes are triggered by JSON "objects". They can map into different
    kinds of Python objects:

    - Plain old dictionaries (with string keys)
    - Dataclasses
    - Type-annotated named tuples

    Other Parameters
    ----------------
    children
        Dictionary of child nodes
    missing_keys
        After fitting, which keys were missing (warning, so far this is tricky
        for unions)
    unwanted_keys
        After fitting, which keys were unwanted (warning, so far this is tricky
        for unions)
    problem_is_kids
        Does the rejection come from the kids? Or is it due to this node only?
    """

    children: Dict[Text, Node] = field(repr=False)
    missing_keys: List[Text] = field(default_factory=list, repr=False)
    unwanted_keys: List[Text] = field(default_factory=list, repr=False)
    problem_is_kids: bool = field(default=False, repr=False)

    def fit_dict(self, t: Type[T]) -> T:
        """
        Fitting a JSON object into a Python dictionary. That's basically just
        a copy with a few sanity checks (and typefitting on the values).

        Notes
        -----
        If we're fitting to a dictionary we're first going to make sure that
        the type spec makes sense and in particular that the keys are strings.
        Indeed, having keys being something else than strings would be much
        more complicated due to the mappings and other transformations
        happening. On the other side, we're doing this for JSON and there is
        no other possibility for keys.

        On the other hand, the values can be fitted into any type that the
        developer asks and a lot of the code is just doing this fitting and
        reporting errors.

        Parameters
        ----------
        t
            Should be a dictionary specification, otherwise it's going to fail
        """

        key_t, value_t = get_args(t)

        if isinstance(key_t, TypeVar):
            key_t = Text

        if not issubclass(key_t, str):
            self.fail("Dictionaries with non-string keys are not supported")

        if isinstance(value_t, TypeVar):
            value_t = Any

        failed = [False]

        def make_dict():
            for k, v in self.children.items():
                try:
                    if not isinstance(k, str):
                        v.fail(f"Key {k!r} is not a string")

                    yield k, self.fitter.fit_node(value_t, v)
                except ValueError:
                    failed[0] = True

        out = dict(make_dict())

        if failed[0]:
            self.problem_is_kids = True
            self.fail("Not all items are fit")

        self.fit_success = True
        return out

    def fit_object(self, t: Type[T]) -> T:
        """
        Fitting into dataclasses and named tuples. On paper that's fairly
        simple, but if you want something safe with good error reporting it's
        a little bit more complicated.

        Notes
        -----
        The first thing that happens is different kinds of inspection on the
        type to figure what are the members of that object, what are their
        types, which ones are mandatory and which ones have a default value,
        etc.

        Then we'll look at each key into the source data and if the data is
        there it'll be fitted into the expected type.

        Please note that there is also some source mapping happening, if the
        dataclass metadata contains a source we're going to use the source
        function instead of directly digging into the value.

        Subsequently, with a few set operations, you can deduce which fields
        are missing, which fields are too much, etc. If anything is wrong, we
        report and fail.

        Finally, the object is instantiated. If the dataclass/named tuple has
        not been tampered with too much, this should pass since we've checked
        that the parameters are right. If there is some funny business going on
        then we're not going to be able to get a decent explanation to the
        user, even potentially crash the whole thing with no explanation. I
        have no better idea on how to handle this though.

        Parameters
        ----------
        t
            Type-annotated named tuple class or dataclass
        """

        sig = signature(t)
        hints = get_type_hints(t)

        kwargs = {}
        fields_sources = {}
        required = set()
        expected = set()
        failed_keys = []

        if is_dataclass(t):
            # noinspection PyDataclass
            for t_field in fields(t):
                if t_field.metadata and "typefit_source" in t_field.metadata:
                    source: Source = t_field.metadata["typefit_source"]
                    fields_sources[t_field.name] = source.value_from_json

        param: Parameter
        for param in sig.parameters.values():
            if param.kind not in {
                Parameter.KEYWORD_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            }:
                continue

            expected.add(param.name)

            if param.default is param.empty:
                required.add(param.name)

            try:
                if param.name in fields_sources:
                    sub_v = fields_sources[param.name](self.children)
                else:
                    sub_v = self.children[param.name]

                if param.name not in hints:
                    sub_v.fail("Missing typing annotations")
            except KeyError:
                pass
            else:
                try:
                    kwargs[param.name] = self.fitter.fit_node(hints[param.name], sub_v)
                except ValueError:
                    failed_keys.append(param.name)

        missing = required - set(kwargs) - set(failed_keys)
        unwanted = set(self.children) - expected
        errors = []

        if missing:
            self.missing_keys = [*missing]
            errors.append(f'Missing keys: {", ".join(repr(x) for x in missing)}')

        if self.fitter.no_unwanted_keys and unwanted:
            self.unwanted_keys = [*unwanted]
            errors.append(f'Unwanted keys: {", ".join(repr(x) for x in unwanted)}')

            for k in unwanted:
                self.children[k].errors.add(f"Unwanted by {format_type_name(t)}")

        if failed_keys:
            errors.append(f'No fit for keys: {", ".join(repr(x) for x in failed_keys)}')

        if errors:
            self.problem_is_kids = True
            self.fail(". ".join([f"Wrong keys set for {format_type_name(t)}", *errors]))

        out = t(**kwargs)
        self.fit_success = True
        return out

    def fit(self, t: Type[T]) -> T:
        """
        This detects if we're dealing with a fit into a dictionary or a fit
        into a named tuple/dataclass.

        Parameters
        ----------
        t
            Type to fit into, either a Dict either NamedTuple either a
            @dataclass
        """

        if get_origin(t) is dict:
            return self.fit_dict(t)
        elif is_named_tuple(t) or is_dataclass(t):
            return self.fit_object(t)
        else:
            self.fail(f"{format_type_name(t)} is not a mapping type")


@dataclass
class ListNode(Node):
    """
    Node that is parent to a list of other nodes. That's basically the
    application of the same type on a list of items instead of just one.

    Other Parameters
    ----------------
    children
        List of child nodes
    problem_is_kids
        Does the rejection come from the kids? Or is it due to this node only?
    """

    children: List[Node] = field(repr=False)
    problem_is_kids: bool = field(default=False, repr=False)

    def fit(self, t: Type[T]) -> T:
        """
        After some sanity checks, goes through the whole list to make sure
        that all the content fits. If some of them don't fit, still continue
        to try fitting the rest, for error reporting purposes.
        """

        if not get_origin(t) is list:
            self.fail(f"{format_type_name(t)} is not a list")

        args = get_args(t)

        if not args:
            self.fail("Could not determine list item type")

        failed = False
        out = []

        for child in self.children:
            try:
                out.append(self.fitter.fit_node(args[0], child))
            except ValueError:
                failed = True

        if failed:
            self.problem_is_kids = True
            self.fail("Not all list items fit")

        self.fit_success = True
        return out


@dataclass
class FlatNode(Node):
    """
    Flat nodes are mostly the builtin types (int, float, str, bool, None) but
    also can be class constructors if the constructor can accept what is being
    passed to it as an argument.
    """

    def fit_builtin(self, t: Type[T]) -> Optional[T]:
        """
        For builtins, we dumbly test one by one. Please note that there is a
        trick where an int can implicitly be converted into a float (since
        it's coming from JSON we know that the int precision is already the one
        of a float so we don't risk on loosing precision).

        Returns None when trying to convert into a type that is not a builtin.

        Parameters
        ----------
        t
            Type to convert into, if not a builtin the None will be returned
        """

        if t is int:
            if not isinstance(self.value, int):
                self.fail(f"Not an int")
            return self.value
        elif t is float:
            if not isinstance(self.value, (int, float)):
                self.fail(f"Neither a float nor an int")
            return float(self.value)
        elif t is str:
            if not isinstance(self.value, str):
                self.fail(f"Not a string")
            return self.value
        elif t is bool:
            if not isinstance(self.value, bool):
                self.fail(f"Not a bool")
            return self.value

    def fit_class(self, t: Type[T]) -> T:
        """
        Fitting the content into a class. That one is a bit tricky but
        basically if the constructor can accept one argument and that this
        argument is properly annotated and that the value can fit into that
        argument's type then the constructor will be called with the fitted
        argument.

        However, so far it only works with builtin types. Arbitrary types might
        come in later, but this is mostly here so that you can transform
        strings into dates.
        """

        if is_dataclass(t) or is_named_tuple(t):
            self.fail(f"{format_type_name(t)} can only fit an object")

        if t is list:
            self.fail(f"{format_type_name(t)} can only fit a list")

        try:
            sig = signature(t).bind(self.value)
        except TypeError:
            self.fail(
                "Constructor should be callable with exactly 1 positional argument"
            )

        (param_name,) = sig.arguments
        param: Parameter = sig.signature.parameters[param_name]

        if param.annotation is param.empty:
            self.fail("Constructor does not specify argument type")

        try:
            arg = self.fitter.fit(param.annotation, self.value)
        except ValueError:
            self.fail(
                f"Constructor {format_type_name(t)} expects {param.annotation} but value does not fit"
            )

        return t(arg)

    def fit(self, t: Type[T]) -> T:
        """
        Tries to use the builtins to fit and if not tries to see if there is
        a constructor we can use.

        Parameters
        ----------
        t
            Type you're fitting into
        """

        if not isclass(t):
            self.fail(f"{t} is not a class")

        out = self.fit_builtin(t)

        if out is None:
            out = self.fit_class(t)

        self.fit_success = True
        return out



from typing import Text

try:
    import pendulum
except ImportError:
    pass
else:
    # noinspection PyInitNewSignature
    class DateTime(pendulum.DateTime):
        """
        Parses a date/time and returns a standard pendulum DateTime.
        """

        def __new__(cls, date: Text) -> pendulum.DateTime:
            self = pendulum.parse(date)

            if not isinstance(self, pendulum.DateTime):
                raise ValueError

            return self

    # noinspection PyInitNewSignature
    class TimeStamp(pendulum.DateTime):
        """
        Parses a Unix timestamp and returns a standard pendulum DateTime.
        """

        def __new__(cls, date: int) -> pendulum.DateTime:
            return pendulum.from_timestamp(date)

    # noinspection PyInitNewSignature
    class Date(pendulum.Date):
        """
        Parses a date and returns a standard pendulum Date.
        """

        def __new__(cls, date: Text) -> pendulum.Date:
            self = pendulum.parse(date)

            if isinstance(self, pendulum.DateTime):
                self = self.date()

            if not isinstance(self, pendulum.Date):
                raise ValueError

            return self



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['typefit']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.12,<0.14', 'pendulum>=2.0,<3.0', 'pygments>=2.6.1,<3.0.0']

setup_kwargs = {
    'name': 'typefit',
    'version': '0.3.0',
    'description': 'Fits JSON values into Python type-anotated objects',
    'long_description': '# TypeFit\n\n[![Read the Docs](https://img.shields.io/readthedocs/typefit)](http://typefit.rtfd.io/)\n[![Build Status](https://img.shields.io/travis/Xowap/typefit)](https://travis-ci.org/Xowap/typefit)\n[![LGTM Grade](https://img.shields.io/lgtm/grade/python/github/Xowap/typefit)](https://lgtm.com/projects/g/Xowap/typefit/)\n\nTyping annotations make Python awesome, however it\'s complicated to keep your\ndata annotated when it comes from external sources like APIs. The goal of\nTypefit is to help you map that external data into type-annotated native Python\nobjects.\n\n```python\nfrom typefit import api\nfrom typing import NamedTuple, Text\n\n\nclass Item(NamedTuple):\n    id: int\n    title: Text\n\n\nclass HackerNews(api.SyncClient):\n    BASE_URL = "https://hacker-news.firebaseio.com/v0/"\n\n    @api.get("item/{item_id}.json")\n    def get_item(self, item_id: int) -> Item:\n        pass\n\nstory = HackerNews().get_item(42)\nprint(story.title)\n# An alternative to VC: &#34;Selling In&#34;\n```\n\nThis is the full example of a Hacker News API client. Its functionality is\nlimited but in 14 lines counting white space you can build a type-safe client\nfor Hacker News. You\'ll find a [full example](example/typefit_hn) attached if\nyou\'re interested.\n\n\n## Documentation\n\n[✨ **Documentation is there** ✨](http://typefit.rtfd.io/)\n\n## Licence\n\nThis library is provided under the terms of the [WTFPL](./LICENSE).\n\nIf you find it useful, you can have a look at the\n[contributors](https://github.com/Xowap/typefit/graphs/contributors) page to\nknow who helped.\n',
    'author': 'Rémy Sanchez',
    'author_email': 'remy.sanchez@hyperthese.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Xowap/typefit/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Callable, Dict, NamedTuple, Optional, Text

Mapping = Dict[Text, Any]


class Source(NamedTuple):
    """
    Provides a way back and forth to convert the data from and to a JSON
    structure. Since the conversion from JSON is able to dig into any number
    of fields from the original mapping, the conversion to JSON will have to
    produce a dictionary as output, even if it has only one key.
    """

    value_from_json: Callable[[Mapping], Any]
    value_to_json: Callable[[Text, Any], Dict]


def meta(source: Optional[Source] = None):
    """
    Generates the field metadata for a field based on what arguments are
    provided. By example, to source a field into a field with another name in
    the JSON data, you can do something like:

    >>> @dataclass
    >>> class Foo:
    >>>     x: int = field(metadata=meta(source=other_field('x')))

    See Also
    --------
    other_field

    Parameters
    ----------
    source
        Source function, that given the mapping as input will provide the
        value as output. If the value isn't found in the mapping, a KeyError
        should arise.
    """

    out = {}

    if source:
        out["typefit_source"] = source

    return out


def other_field(name: Text) -> Source:
    """
    Looks for the value in a field named name.

    Parameters
    ----------
    name
        Name of the field to look for the value into.
    """

    def from_json(mapping: Mapping) -> Any:
        return mapping[name]

    def to_json(field_name: Text, obj: Any) -> Dict:
        return {name: getattr(obj, field_name)}

    return Source(from_json, to_json)
